/* 	ADF4351 Library
	
	Description:
	Class to allow easy control of the ADF4351 synthersizer.
	
	Contributers:
	- Jan Swanepoel, 2018
	
	Notes:
	- Mostly tested on the STM32F103C micro's. On Arduino framework be aware of the 32bit variable limit.
	
	Wiring
	------
	Arduino Uno
	DAT -> Pin 11 (MOSI)
	CLK -> Pin 13 (SCK)
	LE	-> Pin 3
	
	STM32F103C Board
	DAT	-> PA7	(MOSI)
	CLK -> PA5	(SCK)
	LE	-> PA4
	LD	-> PB0
	
	Frequency Resolution
	--------------------
	The frequency resolution is set to allow 1kHz (and even slightly smaller) resolution 
	using RCounter and BandSelectDivider fields. These fields will need to be modified for
	smaller resolutions.
	The resolution is determined by the PFD frequency frac and mod values. For example
	using a 25MHz PFD the smallest fraction we can get is about 25MHz * 1 / 4096 = 6103.52Hz.
	Since the frac and mod values are limited by their 12bit width and the reference 
	oscillator is fixed at 25MHz all we can do is to use the RCounter value to divide the 
	PFD down to a lower frequency. Setting the RCounter to 8 would divide PFD down to 
	3.125MHz and then the smallest fraction would be about 3.125MHz * 1 / 4096 = 762.94Hz. 
	One thing to keep in mind is that the VCO band selection cicruit also uses the output 
	from the RCounter and therefore the BandSelectDivider should also be adjusted to keep 
	it's clock below 125kHz.
	
*/

#include<SPI.h>


class ADF4351
{
	public:
		// Initialize
		void Init();
		
		// Set the output frequency
		void SetFreq(uint64_t freq_Hz);
    
		// Enable or disable RF output. 0 = disable, 1 = enable
		void SetRFOut(uint8_t enabled);
		
		// Set the RF output power
		void SetRFPower(uint8_t power);

		// Get frequency locked states. 0 = not locked, 1 = locked
		bool FreqLocked();
    
		// Update the registers
		void Update();
		
	private:
		struct
		{
			uint16_t Integer;
			uint16_t Fractional;
		} Register0;
		
		struct
		{
			uint8_t PhaseAdjust;
			uint8_t Prescaler;
			uint16_t Phase;
			uint16_t Modulus;
		} Register1;
		
		struct
		{
			uint8_t NoiseMode;
			uint8_t MUXOut;
			uint8_t ReferenceDoubler;
			uint8_t ReferenceDivider;
			uint16_t RCounter;
			uint8_t DoubleBuffer;
			uint8_t ChargePumpCurrent;
			uint8_t LDF;
			uint8_t LDP;
			uint8_t PDPolarity;
			uint8_t PowerDown;
			uint8_t CP;
			uint8_t CounterReset;
		} Register2;
		
		struct
		{
			uint8_t BandSelectClockMode;
			uint8_t ABP;
			uint8_t ChargeCancelation;
			uint8_t CSR;
			uint8_t ClockDividerMode;
			uint16_t ClockDivider;
		} Register3;
		
		struct
		{
			uint8_t FBSelect;
			uint8_t	RFDivider;
			uint8_t BandSelectDivider;
			uint8_t VCOPoweredDown;
			uint8_t MTLD;
			uint8_t AuxOutputSelect;
			uint8_t AuxOutputEnabled;
			uint8_t AuxOutputPower;
			uint8_t RFOutputEnabled;
			uint8_t OutputPower;
		} Register4;
		
		struct
		{
			uint8_t LDPinMode;
		} Register5;
		
		uint32_t r0;
		uint32_t r1;
		uint32_t r2;
		uint32_t r3;
		uint32_t r4;
		uint32_t r5;
		
		uint32_t RFin;			// Reference oscillator frequency
		uint32_t RFin_Er;   // Reference frequency error
		
		uint8_t pin_LE;			// Load Enable pin. Works like chip select
		uint8_t pin_LD;			// Lock Detect pin.
		
		
		// Write to a register on the device
		void WriteRegister(uint32_t data);
		
};


void ADF4351::Init()
{
	pin_LE = PA4;
	pin_LD = PB0;
 
	pinMode(pin_LE, OUTPUT);
	pinMode(pin_LD, INPUT);
	
	SPI.begin();
	SPI.setDataMode(SPI_MODE0);
  
	// Set register 1 default values
	Register1.PhaseAdjust = 0;
	Register1.Prescaler = 1;
	Register1.Phase = 1;
	Register1.Modulus = 0;
	
	// Set register 2 default values
	Register2.NoiseMode = 0;
	Register2.MUXOut = 0;
	Register2.ReferenceDoubler = 0;
	Register2.ReferenceDivider = 0;
	Register2.RCounter = 8;					// Divide RFin by 8 for 1kHz resolution
	Register2.DoubleBuffer = 0;
	Register2.ChargePumpCurrent = 7;
	Register2.LDF = 0;
	Register2.LDP = 0;
	Register2.PDPolarity = 1;
	Register2.PowerDown = 0;
	Register2.CP = 0;
	Register2.CounterReset = 0;
	
	// Set register 4 default values;
	Register4.FBSelect = 1;
	Register4.RFDivider = 1;
	Register4.BandSelectDivider = 25;		// Use value of 200 / RCounter
	Register4.VCOPoweredDown = 0;
	Register4.MTLD = 1;
	Register4.AuxOutputSelect = 0;
	Register4.AuxOutputEnabled = 0;
	Register4.AuxOutputPower = 0;
	Register4.RFOutputEnabled = 0;
	Register4.OutputPower = 3;

	RFin_Er = 500;							// Adjust to reference frequency error
	RFin = 25e6 - RFin_Er;
	
	Update();
}


void ADF4351::SetFreq(uint64_t freq_Hz)
{
	double Fpfd = RFin / 8;		// Phase Frequency Detector frequency, in other words the Phase detector input frequency.
	
	// First step is to get the RF output divider
	if (freq_Hz >= 2200000000)
		Register4.RFDivider = 0;		// Divide by 1

	if (freq_Hz < 2200000000)
		Register4.RFDivider = 1;		// Divide by 2

	if (freq_Hz < 1100000000)
		Register4.RFDivider = 2;		// Divide by 4

	if (freq_Hz < 550000000)
		Register4.RFDivider = 3;		// Divide by 8

	if (freq_Hz < 275000000)
		Register4.RFDivider = 4;		// Divide by 16

	if (freq_Hz < 137500000)
		Register4.RFDivider = 5;		// Divide by 32

	if (freq_Hz < 68750000)
		Register4.RFDivider = 6;		// Divide by 64
	
	// Calculate the VCO output frequency using the selected divider
	double VCO_Freq = freq_Hz * pow(2, Register4.RFDivider);

	// Calculate the PLL integer divider
	Register0.Integer = VCO_Freq / Fpfd;

	// Calculate MOD for 1kHz resolution
	Register1.Modulus = Fpfd / 1e3;
	
	// Calculate the Fractional value with FRAC = MOD * ((VCOout/Fpfd) - INT)
	Register0.Fractional = Register1.Modulus * (((float)VCO_Freq / Fpfd) - (float)Register0.Integer);

/*	
	Serial.print("VCO: ");
	Serial.println(VCO_Freq);
	
	Serial.print("INT: ");
	Serial.println(Register0.Integer);

	Serial.print("MOD: ");
	Serial.println(Register1.Modulus);

	Serial.print("FRAC: ");
	Serial.println(Register0.Fractional);
	*/
}


void ADF4351::SetRFOut(uint8_t enabled)
{
	if (enabled == 0)
		Register4.RFOutputEnabled = 0;
	else
		Register4.RFOutputEnabled = 1;
}


void ADF4351::SetRFPower(uint8_t power)
{
	// Disgard invalid power values
	if (power > 3)
		return;
	
	Register4.OutputPower = power;
}


void ADF4351::Update()
{
	//todo:  put masks on the fields...
	
	// Build register 0
	r0 = (uint32_t)(Register0.Integer) << 15 |
		 (uint32_t)(Register0.Fractional) << 3 |
		 0x00000000;
	
	// Build register 1
	r1 = (uint32_t)(Register1.PhaseAdjust) << 28 |
		 (uint32_t)(Register1.Prescaler) << 27 |	
		 (uint32_t)(Register1.Phase) << 15 |
		 (uint32_t)(Register1.Modulus) << 3 |
		 0x00000001;
	
	// Build register 2
	r2 = (uint32_t)(Register2.NoiseMode) << 29 |
		 (uint32_t)(Register2.MUXOut) << 26 |
		 (uint32_t)(Register2.ReferenceDoubler) << 25 |
		 (uint32_t)(Register2.ReferenceDivider) << 24 |
		 (uint32_t)(Register2.RCounter) << 14 |
		 (uint32_t)(Register2.DoubleBuffer) << 13 |
		 (uint32_t)(Register2.ChargePumpCurrent) << 9 |
		 (uint32_t)(Register2.LDF) << 8 |
		 (uint32_t)(Register2.LDP) << 7 |
		 (uint32_t)(Register2.PDPolarity) << 6 |
		 (uint32_t)(Register2.PowerDown) << 5 |
		 (uint32_t)(Register2.CP) << 4 |
		 (uint32_t)(Register2.CounterReset) << 3 |
		 0x00000002;
	
	r3 = 0x000004B3;
	
	// Build register 4
	r4 = (uint32_t)(Register4.FBSelect) << 23 |
		 (uint32_t)(Register4.RFDivider) << 20 |
		 (uint32_t)(Register4.BandSelectDivider) << 12 |
		 (uint32_t)(Register4.VCOPoweredDown) << 11 |
		 (uint32_t)(Register4.MTLD) << 10 |
		 (uint32_t)(Register4.AuxOutputSelect) << 9 | 
		 (uint32_t)(Register4.AuxOutputEnabled) << 8 | 
		 (uint32_t)(Register4.AuxOutputPower) << 6 | 
		 (uint32_t)(Register4.RFOutputEnabled) << 5 | 
		 (uint32_t)(Register4.OutputPower) << 3 | 
		 0x00000004;
	
	r5 = 0x00580005;
	
	// Write the registers	 
	WriteRegister(r0);
	WriteRegister(r1);
	WriteRegister(r2);
	WriteRegister(r3);
	WriteRegister(r4);
	WriteRegister(r5);
	
/*	
	Serial.println(r0, HEX);
	Serial.println(r1, HEX);
	Serial.println(r2, HEX);
	Serial.println(r3, HEX);
	Serial.println(r4, HEX);
	Serial.println(r5, HEX);*/
}


bool ADF4351::FreqLocked()
{
	return digitalRead(pin_LD);
}


void ADF4351::WriteRegister(uint32_t data)
{
	digitalWrite(pin_LE, LOW);

	// Writes the 32bit data bytes
	for(int i = 0; i < 4 ; i++)
	{
		uint8_t dataByte = data>>(8 * (3-i));
		SPI.transfer(dataByte);
	}

	digitalWrite(pin_LE, HIGH);
}
