/*  ADF Based Signal Generator Firmware
 *  Jan Swanepoel, 2018 
 *  
*/
#include "ADF4351.h"
#include "scpi.h"


SCPI cmdInterface;
ADF4351 sigGen;

// Global variables
bool sweeping;
double sweepStartFreq;
double sweepStopFreq;
double sweepStep;
double sweepCurrentFreq;


// Set the output frequency.
// Parameter is frequency in Hz 
float SetFixedFrequency(double freq)
{
  uint64_t ifreq = (uint64_t)freq;
   
  Serial.print("Set frequency: ");
  Serial.println(ifreq);
  
  sigGen.SetFreq(ifreq);
  sigGen.Update();
}


// Set RF output power in 3dB steps. 
// Parameter is a value from 0 to 3
float SetRFPower(double pwr)
{
  uint8_t ipwr = (uint8_t)pwr;

  Serial.print("Set Power: ");
  Serial.println(ipwr);

  sigGen.SetRFPower(ipwr);
  sigGen.Update();
}


// Enable or diaable RF output
// Parameter is 1 for enable and 0 for disable
float SetRFOut(double rfout)
{
  uint8_t irfout = (uint8_t)rfout;
  
  if (irfout > 0)
  {
    sigGen.SetRFOut(1);
    Serial.println("RF ON!");
  }
  else
  {
    sigGen.SetRFOut(0);
    Serial.println("RF OFF!");
  }
  
  sigGen.Update();
}


// Determine if the PLL is locked
// Takes no parameter
float GetFreqLock(double na)
{
  if (sigGen.FreqLocked() > 0)
    Serial.println("Locked");
  else
    Serial.println("No Lock");  
}


// Start or stop a sweep operation
// Parameter is 1 for start and 0 for stop 
float Sweep(double startSweep)
{
  if (startSweep > 0)
  {
    sweeping = true;
    sweepCurrentFreq = sweepStartFreq;
  }
  else
    sweeping = false;
}


// Set the start frequency used for sweeping
// Parameter is the frequency in Hz
float SetSweepStart(double freqStart)
{
  sweepStartFreq = freqStart;
}


// Set the stop frequency used for sweeping
// Parameter is the frequency in Hz
float SetSweepStop(double freqStop)
{
  sweepStopFreq = freqStop;
}


// Initialize
void setup() 
{
  Serial.begin(57600);
  Serial.println("Starting...");

  pinMode(PA0, INPUT_ANALOG);
  
  // Set default sweep settings
  sweepStartFreq = 35e6;
  sweepStopFreq = 4.4e9;
  sweepStep = 1e6;
  
  sigGen.Init();
  
  uint8_t grpConfig = cmdInterface.CreateGroup("CONFIG", 0);
  uint8_t grpSource = cmdInterface.CreateGroup("SOURCE", 0);

  cmdInterface.RegisterParameter("FREQ", grpSource, &SetFixedFrequency);
  cmdInterface.RegisterParameter("PWR", grpSource, &SetRFPower);
  cmdInterface.RegisterParameter("RFON", grpSource, &SetRFOut);
  cmdInterface.RegisterParameter("FREQLOCK?", grpSource, &GetFreqLock);
  cmdInterface.RegisterParameter("SWEEP", grpSource, &Sweep);
  cmdInterface.RegisterParameter("SWEEPSTART", grpSource, SetSweepStart);
  cmdInterface.RegisterParameter("SWEEPSTOP", grpSource, SetSweepStop);

  Serial.println(cmdInterface.GetGroupName(grpSource));
  Serial.println(cmdInterface.GetGroupID("SOURCE"));

}

// Main loop
void loop() 
{
  if (Serial.available() > 0)
    cmdInterface.Parse(Serial.read());


  if (sweeping && sigGen.FreqLocked())
  {
    sweepCurrentFreq += sweepStep;
    
    if (sweepCurrentFreq < sweepStopFreq)
    {
      SetFixedFrequency(sweepCurrentFreq);
    }
    else
      sweepCurrentFreq = sweepStartFreq;
  }


}
