/* 	SCPI (Standard Commands for Programmable Instruments)
*	Jan Swanepoel,
*	2018
*
*/

#define PARAM_MAX		10
#define GROUP_MAX		10
#define CMD_LEN_MAX		30

class SCPI
{
	public:
		// Define function pointer type that takes a float argument and returns a float
		typedef float (*func_t)(double);
		
		SCPI();
		
		uint8_t CreateGroup(const char* name, uint8_t parent);
		
		const char* GetGroupName(uint8_t index);
		
		uint8_t GetGroupID(char* name);
		
		uint8_t RegisterParameter(char* command, uint8_t group, func_t function);
		
		uint8_t GetParameterID(char* name);
		
		void Parse(char byte);
	
	
	private:
		
		struct GroupType
		{
			uint8_t id;
			uint8_t parentId;
			const char* name;
		};
		
		struct ParameterType 
		{
			uint8_t id;			// Parameter ID. Should be larger than 0, value of 0 means unused.
			uint8_t groupId;
			char* name;
			func_t function;
		};
		
		uint8_t grpIndex;
		
		GroupType groups[GROUP_MAX];
		ParameterType Parameters[PARAM_MAX];
		
		char cmdBuilder[CMD_LEN_MAX];
		uint8_t cmdBuilderIndex;
		
		bool SubString(char* string, char startChar, char endChar, char *substring);

};


SCPI::SCPI()
{
	grpIndex = 1;
	cmdBuilderIndex = 0;
}


uint8_t SCPI::CreateGroup(const char* name, uint8_t parent)
{
	groups[grpIndex].id = grpIndex;
	groups[grpIndex].name = name;
	
	grpIndex++;
	
	return grpIndex - 1;
}


const char* SCPI::GetGroupName(uint8_t index)
{
	return groups[index].name;
}


uint8_t SCPI::GetGroupID(char* name)
{
	for (int i = 0; i < GROUP_MAX; i++)
		if (strcmp(name, groups[i].name) == 0)
			return groups[i].id;
		
	return 0;
}


uint8_t SCPI::RegisterParameter(char* command, uint8_t group, func_t function)
{
	uint8_t newId = 1;
	
	// Find first available id that can be used
	for (int i = 0; i < PARAM_MAX; i++)
		if (Parameters[i].id == newId)
		{
			newId++;
			i = 0;
		}

		
	for (int i = 0; i < PARAM_MAX; i++)
		if (Parameters[i].id == 0)
		{
			Parameters[i].id = newId;
			Parameters[i].groupId = group;
			Parameters[i].name = command;
			Parameters[i].function = function;
			break;
		}
	
	return newId;
}


uint8_t SCPI::GetParameterID(char* name)
{
	for (int i = 0; i < PARAM_MAX; i++)
		if (strcmp(name, Parameters[i].name) == 0)
			return Parameters[i].id;
		
	return 0;
}


void SCPI::Parse(char c)
{
	cmdBuilder[cmdBuilderIndex] = c;
	cmdBuilderIndex++;
  
	if ((c == '\r') || (c == ';'))
	{
		char group[10];
		char command[16];
		char paramValue[12];
		
		if (!SubString(cmdBuilder, ':', ':', group))
		{
			Serial.println("Invalid group");
			cmdBuilderIndex = 0;
			return;
		}
		
		bool cmdSet = SubString(cmdBuilder, ':', '\r', command);
		bool cmdSetParam = SubString(cmdBuilder, ':', ' ', command);
		
		if (!(cmdSet || cmdSetParam))
		{
			Serial.println("Invalid command");
			cmdBuilderIndex = 0;
			return;
		}
		
		if (cmdSetParam)
			SubString(cmdBuilder, ' ', '\r', paramValue);	
			
		uint8_t grpId = GetGroupID(group);
		uint8_t paramId = GetParameterID(command);
				
		if (paramId > 0)
		{
			for (int i = 0; i < PARAM_MAX; i++)
				if ((Parameters[i].groupId == grpId) && (Parameters[i].id == paramId))
				{					
					float v = strtod(paramValue, NULL);
					Parameters[i].function(v);
				}
		}	
		else
			Serial.println("Invalid parameter");
		
		// Reset commands builder index
		cmdBuilderIndex = 0;
	}
	
}



bool SCPI::SubString(char* string, char startChar, char endChar, char *substring)
{
	uint8_t stringLen = 0;
	uint8_t startCharIndex = 0;
	uint8_t endCharIndex = 0;
	
	// Get string length
	for (int s = 0; s < 255; s++)
	{
		if (string[s] == endChar)
			endCharIndex = s;
		
		if (string[s] == '\r')
		{
			stringLen = s;
			break;
		}
	}
	
	for (int s = 0; s < stringLen; s++)	
		if ((string[s] == startChar) && (s != endCharIndex))	// Check indices are different to allow same separation character 
			startCharIndex = s;
	
	// Simple sanity check
	if (endCharIndex == 0)
		return false;
	
	// Copy substring
	int n;
	for(n = 0; n < (endCharIndex - startCharIndex) - 1; n++)
		substring[n] = string[startCharIndex + 1 + n];
	
	// End string
	substring[n] = '\0';
	
	return true;
}
